package HTTP::StegTest;
our $VERSION=0.01;

=head1 NAME

HTTP::StegTest

=head1 DESCRIPTION

Recover and locally store images from the web, 
including those linked by anchor, image map, and frames.

Record any differences in file sizes between probes, 
and scan all images with an external tool 
(currently supports stegdetect).

Log all activity and create html pages to report information, 
both rolled up and seperated between sites.


=head1 SYNOPSIS

	use HTTP::StegTest;
	new HTTP::StegTest (
		'/image/save/dir',
		['http://www.getthese.com/all','http://get.this/'],
		['http://www.somewhere/ignorethis.html','http://and.this.html']
	);

	new HTTP::StegTest (
 		'/image/commondir/dir',
		['http://www.getthese.com/all','http://get.this/'],
		['NONE'],
		['http://www.getthese.com/minimumURLtoMatch'],
		$minsize,
		'/path/to/sd',
		'/image/commondir'
	);


"new" does all the steps without prompts if all fields are used.

	$blah = HTTP::StegTest::stegtest (
		'/image/commondir/dir',
		'/image/commondir',
		'/path/to/sd',

	);
	
"stegtest" subroutine skips fetching images and scans 
all images in dir with sd, It also does the reporting step.

	$blah = HTTP::StegTest::report (
		'/image/commondir/dir',
		'/image/commondir',
	);

"report" subroutine only writes html reports from the 
log files produced by the other steps.

To view reports, open the index.html document created in the 
images directory (or the common directory).

=head1 DEPENDENCIES

	strict;
	warnings;
	Carp;
	LWP::UserAgent;
	HTTP::Request;
	HTML::TokeParser;

=cut

use LWP::UserAgent;
use HTTP::Request;
use HTML::TokeParser;
use Carp;
use strict;
use warnings;
no strict 'refs';

=head1 PACKAGE GLOBAL VARIABLE

=head2 $CHAT

Set to above zero if you'd like a real-time report to C<STDERR>.
Defaults to off.

=cut

our $CHAT = 0;
$CHAT =0;

=head2 $EXTENSIONS_RE, $EXTENSIONS_BAD

A regular expression 'or' list of image extensions to match.

Will be applied at the end of a filename, after a point, and is insensitive to case.

Defaults to C<(jpg|jpeg|bmp|gif|png|xbm|xmp|doc)>.

$EXTENSIONS_BAD are files to skip parsing, such as 10 MB avi files. 
All filenames which match will be skipped, but recorded to MEDIA.txt.

Defaults to (wmv|avi|rm|mpg|asf|ram|asx|mpeg|mp3|swf)'

=cut

our $EXTENSIONS_RE = '(jpg|jpeg|bmp|gif|png|xbm|xmp|doc)';
our $EXTENSIONS_BAD = '(wmv|avi|rm|mpg|asf|ram|asx|mpeg|mp3|swf)';

=head2 $CMP

When filled with the path to an executable such as cmp, it will run with 2 versions of a changed file.

Output will be recorded to CHANGED.txt file, Meant to record by-byte differences.

Make equal to 'NONE' to skip altogether

=cut

our $CMP = "e:\\images\\tools\\cmp.exe";

=head2 $NEWNAMES

Set to above zero to save files with new names; defaults to using original names; 1 appends timestamnp; 2 is URL in name.

=cut

our $NEWNAMES = 2;

=head2 $HTML_HEADRE, $HTML_FOOTER

Used for the reports, edit them at will. Future version may allow this to be external.

=cut

our $HTML_HEADER ="<html><head><title>Steg Test Output</title></head>
<body><br><br><center><b>Results</b></center><br><br>";

our $HTML_FOOTER ="<br><br><center><font size=1><i><a href=http://www.duncanlamb.com/stegtest/>Stegtest</a>, Copyright 2001 Duncan Lamb</i></font></center></body></html>";

=head1 CONSTRUCTOR METHOD new

Besides the class reference, accepts:

=item 1.

the path to the directory in which to store images (no trailing oblique necessary);

=item 2.

reference to array of URLs to process;

=item 3.

reference to array of URLs to ignore.

If one of these is C<ALL>, then will ignore all B<HTML> documents not in the referenced array of URLs to process.
If one of these is C<NONE>, will ignore no documents.

Returns a blessed hash, keys of which are:

=item 4.

The minimum path the URL must contain.

=item 5.

The minimum size an image must be to be saved.

=item DONE

a hash keys of which are the original URLs of the images, value being are the local filenames.

=item FAILED

a hash, keys of which are the failed URLs, values being short reasons.

=cut


sub new { my ($class,$dir,$dodo,$dont,$MINURL,$MINIMGSIZE,$SDETECT,$commondir) = (shift,shift,shift,shift,shift,shift,shift,shift);
	warn "Usage: \$class::new (\$dir,\\\@do,\\\@dont) " and return undef if not defined $class;
	warn "$class::new requires a directory in which store the images as its first argument." and return undef if not defined $dir;
	warn "$class::new requires at least one URL to saerch for images as its seoncd argument." and return undef if not defined $dodo;

	my $self={};
	$self->{DONE} = {};
	$self->{FAILED} = {};
	bless $self,$class;

	$self->{MINIMGSIZE} = $MINIMGSIZE if defined $MINIMGSIZE;
	warn "Minimum image size defined as $self->{MINIMGSIZE}.\n" if $CHAT;

	my @minurls;
	my $muflag;
	
	my @lines;

	if (defined $MINURL){
		foreach (@$MINURL){ push @minurls,$_; 
		warn "URL must begin with $_ ..." if $CHAT; }
	}

#	$self->{MINURL} = $MINURL if defined $MINURL;
#	warn "URL must begin with $MINURL ..." if $CHAT;

	my @urls;
	my @aurls;

	foreach (@$dodo){ push @urls,$_; $self->{DO}->{$_}=1; }
	undef $dodo;
	foreach (@$dont){ $self->{IGNORE}->{$_} = '1' }
	undef $dont;
	
	@aurls = @urls;
# set up logging directories, etc
	
	if (not (-d $dir."/DIFFERENCES")){mkdir $dir."/DIFFERENCES";}
	our $logdir = $dir."/LOGS";
	if (not (-d $logdir)){mkdir $dir."/LOGS";}
	if (not (-d $dir."/POSITIVES")){mkdir $dir."/POSITIVES";}

	my $ctime;
	$ctime = localtime(time);  
	open LINK,">>$logdir/log.txt" or warn "Couldn't open to save <logfile>!" and return "Failed to save.";
	print LINK "$dir $ctime\n";
	close LINK;
	open LINK,">>$logdir/common.txt" or warn "Couldn't open to save <logfile>!" and return "Failed to save.";
	foreach (@aurls){
		print LINK "$dir $_\n";
	}
	close LINK;
	if (defined $commondir and $dir ne $commondir){
		open LINK,">>$commondir/log.txt" or warn "Couldn't open to save <logfile>!" and return "Failed to save.";
		print LINK "$dir $ctime\n";
		close LINK;
		open LINK,">>$commondir/common.txt" or warn "Couldn't open to save <logfile>!" and return "Failed to save.";
		foreach (@aurls){
			print LINK "$dir $_\n";
		}
		close LINK;
	}

# main:

	DOC:
	while (my $doc_url = shift @urls){

		warn "-"x60,"\n" if $CHAT;
		my ($doc,$p);
		if (exists $minurls[0]){
			$muflag =1;
			foreach (@minurls){
				if ($doc_url =~ /^$_/){
					warn "URL matches scope, $doc_url $_\n" if $CHAT;
					$muflag = 0;
				}
			if ($muflag){next DOC;}
			}
		}

		if (exists $self->{FAILED}->{$doc_url} or exists $self->{DONE}->{$doc_url}){
			warn "Already done $doc_url.\n" if $CHAT;
			next DOC;
		}

		if (exists $self->{IGNORE}->{$doc_url}){
			warn "In IGNORE list: $doc_url.\n" if $CHAT;
			next DOC;
		}

		# Not in do list, not an image, not run with IGNORE NONE option
		if (not exists $self->{DO}->{$doc_url} and $doc_url !~ m|(\.$EXTENSIONS_RE)$|i
		and not exists $self->{IGNORE}->{NONE}){
			warn "Not in DO list - ignoreing $doc_url .\n" if $CHAT;
			$self->{IGNORE}->{$doc_url} = "Ignoring";
			next DOC;
		}

		unless ($doc = $self->get_document($doc_url,$dir,$commondir)){
			$self->{FAILED}->{$doc_url} = "Agent couldn't open page";
			next DOC;
		}

		# If an image, save it
		if ($doc_url =~ m|(\.$EXTENSIONS_RE)$|i) {
			$self->{DONE}->{$doc_url} = $self->_save_img($dir,$doc_url,$doc,$commondir);
			next DOC;
		} else {
			$self->{DONE}->{$doc_url} = "Did HTML.";
		}

		# Otherwise try to parse it
		unless ($p = new HTML::TokeParser( \$doc )){
			warn "* Couldn't create parser from \$doc\n" if $CHAT;
			$self->{FAILED}->{$doc_url} = "Couldn't create agent parser";
			next DOC;
		}
		warn "OK - parsing document $doc_url ...\n" if $CHAT;

		while (my $token = $p->get_token){
			if (@$token[1] eq 'img' and defined @$token[2]->{src}){
				warn "*** Found image: @$token[2]->{src}\n" if $CHAT;
				my $uri = &abs_url( $doc_url, @$token[2]->{src} );

				if (not exists $self->{IGNORE0}->{$uri} and not exists $self->{DONE}->{$uri} and not exists $self->{FAILED}->{$uri}){
					unshift @urls, $uri;
				}
			}
			elsif (@$token[1] eq 'frame' and defined @$token[2]->{src}){
				warn "*** Found frame: @$token[2]->{src}\n" if $CHAT;
				my $uri = &abs_url( $doc_url, @$token[2]->{src} );
				if (not exists $self->{IGNORE0}->{$uri} and not exists $self->{DONE}->{$uri} and not exists $self->{FAILED}->{$uri}){
					unshift @urls, $uri;
				}
			}
			elsif ((@$token[1] eq 'area' or @$token[1] eq 'a' and @$token[0] eq 'S') and defined @$token[2]->{href}){
				warn "*** Found link: @$token[2]->{href}\n" if $CHAT;
				my $uri = &abs_url( $doc_url, @$token[2]->{href} );
				if (not exists $self->{IGNORE}->{$uri} and not exists $self->{DONE}->{$uri} and not exists $self->{FAILED}->{$uri}){
					unshift @urls, $uri;
					if (@$token[1] eq 'a' and @$token[0] eq 'S' and $uri =~ /extreme/){
						open LINK,">>$logdir/TRACKERS.txt" or warn "Couldn't open to save <TRACKERS>!" and return "Failed to save.";
							print LINK "$uri\n";
						close LINK;
						if (defined $commondir and $dir ne $commondir){
							open LINK,">>$commondir/TRACKERS.txt" or warn "Couldn't open to save <TRACKERS>!" and return "Failed to save.";
								print LINK "$uri\n";
							close LINK;
						}
					}
					if (@$token[1] eq 'a' and @$token[0] eq 'S' and length($uri) > 300){
						open LINK,">>$logdir/TOOLONG.txt" or warn "Couldn't open to save <TOOLONG>!" and return "Failed to save.";
							print LINK "$uri\n";
						close LINK;
						if (defined $commondir and $dir ne $commondir){
							open LINK,">>$commondir/TOOLONG.txt" or warn "Couldn't open to save <TOOLONG>!" and return "Failed to save.";
							print LINK "$uri\n";
							close LINK;
						}

					}
				}
			}
		}

	} # Next DOC

	stegtest($dir,$commondir,$SDETECT,1);
	report($dir,$commondir);

	return $self;
}	#start edit	


sub stegtest {my ($dir, $commondir, $SDETECT, $rflag) = (shift, shift, shift, shift);

	my @lines;
	my $logdir = $dir."/LOGS";
	if (not (-d $logdir)){mkdir $logdir;}
	if (defined $SDETECT){
	
		@lines = `$SDETECT $dir/*.jpg`;	
		if (defined $commondir and $dir ne $commondir){
			open FP, ">>$commondir/FALSEPOSITIVES.txt";
			open POS, ">>$commondir/POSITIVES.txt";
		}
		open FPL, ">>$logdir/FALSEPOSITIVES.txt";
		open POSL, ">>$logdir/POSITIVES.txt";
		
		foreach (@lines){
		
			if ($_ =~ / : negative$/){
				$_ = $_;
	 		} elsif ($_ =~ /false positive likely/) {
				print FPL "$_";
				if (defined $commondir and $dir ne $commondir){print FP "$_";}
	 		} else {
				print POSL "$_";
				# need to add some code in to copy images to a "positives" directory to isolate them
				if (defined $commondir and $dir ne $commondir){print POS "$_";}
	 		}
	
		}
		warn @lines if $CHAT;
		if (defined $commondir and $dir ne $commondir){
			close FP;
			close POS;
		}
		close FPL;
		close POSL;
	
	
	}  # end stegdetect, uses $dir, $commondir, declare @lines
	if (not defined $rflag) {report($dir,$commondir);} 	#run only if invoked directly

}
  # start reporting, uses $dir, $commondir, $logdir, declare @lines
	
sub report{ my ($dir, $commondir) = (shift, shift);	# home page
	my $logdir = $dir."/LOGS";
	my @lines;
	my $i;
	my $x;
	my %sites;
	my %alreadydone;
	my $tdir = $dir;
	my $tcommondir;
	my $di;
	my @dirs;
	my $udirs;
	$dirs[0] = $dir;
	if (defined $commondir and $dir ne $commondir){
		$tcommondir = $commondir;
		$tcommondir =~ s/\//\/\//g;
		$tcommondir =~ s/\\/\\\\/g;	
		$dirs[1] = $commondir;
	}
	$tdir =~ s/\//\/\//g;
	$tdir =~ s/\\/\\\\/g;
	

	if (defined $commondir and $dir ne $commondir){
		open (COUT, ">$commondir/index.html");
		print COUT $HTML_HEADER;
		print COUT "Sites scanned (click for local images and reports):<p><blockquote>";
		open (IN,"<$commondir/common.txt");
			@lines = <IN>;
		close IN;
		foreach (@lines){
			$_ =~ / /;
			$sites{$`} = 1;	#only store directories once
		}
		foreach (keys %sites){			
			open (IN,"<$_/LOGS/common.txt") or die "NOT THERE";
				@lines = <IN>;
			close IN;
			foreach (@lines){		
				$_ =~ / /;
				$x = $`; $i = $';
				$x =~ s/$tcommondir//;
				$x =~ s/^(\/|\\)//;
				
				if (exists $alreadydone{$i}){next;}	#only print URL once	
				print COUT "<a href=$x/index.html>$i</a><br>";
				$alreadydone{$i} = $x;
			
			}
			print COUT "<p>";
		
		}
		undef %alreadydone;
		print COUT "</blockquote><b>Combined Stats:</b><p>";
	}

	open (OUT, ">$dir/index.html"); 
	print OUT $HTML_HEADER;
	
	$_ = $lines[$#lines-1];
	/ /;
	print OUT "Last run for $dir at $'<p>";	#date
	print OUT "For Sites: <p>\n";

	open (IN,"<$logdir/common.txt");
		@lines = <IN>;
	close IN;
	
	
	
	
	open (IN,"<$logdir/NEW.txt");
		@lines = <IN>;
	close IN;
	print OUT "<a href=imagelist.html>Total Images:</a> ".@lines."<br>\n";
	open (EOUT,">$dir/imagelist.html");
		print EOUT $HTML_HEADER;
		print EOUT "<b><center>Image Files grabbed</center></b><p>";
		print EOUT "<table><tr><td align=center><b>URL</b></td><td align=center><b>Local Copy</b></td></tr>";	
		foreach (@lines){
			print EOUT "<tr><td><font size=2><a href=$_>$_</a></td><td><font size=2>";
			$_ =~ s/\/|\:|\~/_/g;			
			print EOUT "<a href=$_>$_</a></td></tr>\n";
		}
		print EOUT "</table>$HTML_FOOTER";			
	close EOUT;	
	
	foreach $di (@dirs){

	if ($di eq $dirs[0]){ $i = $logdir; $udirs = $tdir;} else {$i = $di; $udirs = $tcommondir}
	if (-f "$i/CHANGED.txt"){

		undef @lines;
		open (IN,"<$i/CHANGED.txt");
			@lines = <IN>;
		close IN; 
		$x = 0;
		open (EOUT,">$di/changed.html");
			print EOUT $HTML_HEADER;
			print EOUT "<b><center>Images which have changed between probes</center></b><p>\n";
			print EOUT "<table><tr><td align=center><b>Old Image</b></td><td align=center><b>New Image</b></td><td align=center><b>Byte Comparison</b></td></tr>\n";	
			for ($i=0; $i < (@lines-1); $i++) {
				if ($lines[$i] =~ /($udirs)/){ 
					$x++;
					$_ = $lines[$i];
					$_ =~ s/($udirs)//;
					$_ =~ s/^(\/|\\)//;
					$_ =~ m/(\w|\.|\/|\\|-)+/;
					print EOUT "<tr><td><center><img src=$&><br><font size=1>$_ bytes</td>";

					$i++;
					$_ = $lines[$i];
					$_ =~ s/($udirs)//;
					$_ =~ s/^(\/|\\)//;
					$_ =~ m/(\w|\.|\/|\\|-)+/;					
					print EOUT "<td><center><img src=$&><br><font size=1>$_ bytes</td>";
					
					print EOUT "<td><textarea width = 20 rows=10>";
					$i++;
					while (exists $lines[$i] and ($lines[$i] !~ m/($udirs)/) and ($lines[$i] !~ m/^\n/)){
						print EOUT "$lines[$i]";
						$i++;
					}

					print EOUT "</textarea></td></tr>\n";
					#$i--;
				}	



			}
			print EOUT "</table>$HTML_FOOTER";			
		close EOUT;	
		if ($di eq $dirs[0]){
			print OUT "<a href=changed.html>Changed Images:</a> ".$x."<br>\n";
		} else {
			print COUT "<a href=changed.html>Changed Images:</a> ".$x."<br>\n";
		}	
		$x=0;
	} else {
		if ($di eq $dirs[0]){
			print OUT "No Changed Images<br>\n";
		} else {
			print COUT "No Changed Images<br>\n";
		}	
		
	}
	}

	foreach $di (@dirs){
	if ($di eq $dirs[0]){ $i = $logdir; $udirs = $tdir;} else {$i = $di; $udirs = $tcommondir}
	if (-f "$i/FALSEPOSITIVES.txt"){
		open (IN,"<$i/FALSEPOSITIVES.txt");
			@lines = <IN>;
		close IN;
		
		open (EOUT,">$di/falsepos.html");
			print EOUT $HTML_HEADER;
			print EOUT "<b><center>Probable False Positives</center></b><p>";
			print EOUT "<table><tr><td align=center><b>Image</b></td><td align=center><b>Local Copy</b></td></tr>";	
			foreach (@lines){
				if ($_ =~ /^\n$/){next;}
				$_ =~ s/($udirs)//;
				$_ =~ s/^(\/|\\)//;
				$_ =~ m/(\w|\.|\/|\\|-)+/;
								
				if (exists $alreadydone{$_}){next;}
				print EOUT "<tr><td>$_</td><td><font size=2>";			
				print EOUT "<img src=$&></td></tr>\n";
				$alreadydone{$_} =1;
			}
			print EOUT "</table>$HTML_FOOTER";			
		close EOUT;
		$_ = keys %alreadydone;
		if ($di eq $dirs[0]){
			print OUT "<a href=falsepos.html>False Positives:</a> $_<br>\n";
		} else {
			print COUT "<a href=falsepos.html>False Positives:</a> $_<br>\n";
		}
		undef %alreadydone;
		
	} else {
		print OUT "No Images Tested Positive<br>\n";
	}
	}

	foreach $di (@dirs){
	if ($di eq $dirs[0]){ $i = $logdir; $udirs = $tdir;} else {$i = $di; $udirs = $tcommondir}
	if (-f "$i/POSITIVES.txt"){
		open (IN,"<$i/POSITIVES.txt");
			@lines = <IN>;
		close IN;
		
		open (EOUT,">$di/pos.html");
			print EOUT $HTML_HEADER;
			print EOUT "<b><center>Positive Matches</center></b><p>";
			print EOUT "<table><tr><td align=center><b>Image</b></td><td align=center><b>Local Copy</b></td></tr>";	
			foreach (@lines){
				if ($_ =~ /^\n$/){next;}
				$_ =~ s/($udirs)//;
				$_ =~ s/^(\/|\\)//;
				$_ =~ m/(\w|\.|\/|\\|-)+/;

				if (exists $alreadydone{$_}){next;}
				print EOUT "<tr><td>$_</td><td><font size=2>";			
				print EOUT "<img src=$&></td></tr>\n";
				$alreadydone{$_} =1;
			}
			print EOUT "</table>$HTML_FOOTER";			
		close EOUT;
		$_ = keys %alreadydone;
		if ($di eq $dirs[0]){
			print OUT "<a href=pos.html>Images Positive during Steganography Detection:</a> ".$_."<br>\n";
		} else {
			print COUT "<a href=pos.html>Images Positive during Steganography Detection:</a> ".$_."<br>\n";
		}		
		undef %alreadydone;

	} 
	}

	foreach $di (@dirs){
	if ($di eq $dirs[0]){ $i = $logdir; $udirs = $tdir;} else {$i = $di; $udirs = $tcommondir}
	if (-f "$i/MEDIA.txt"){
		open (IN,"<$i/MEDIA.txt");
			@lines = <IN>;
		close IN;
		if ($di eq $dirs[0]){
			print OUT "<a href=media.html>A/V Media Files:</a> ".@lines."<br>\n";
		} else {
			print COUT "<a href=media.html>A/V Media Files:</a> ".@lines."<br>\n";
		}
		
		
		open (EOUT,">$di/media.html");
			print EOUT $HTML_HEADER;
			print EOUT "<b><center>Media Files</center></b><p>";	
			foreach (@lines){
				print EOUT "<a href=$_>$_</a><br>\n";
			}
			print EOUT $HTML_FOOTER;			
		close EOUT;
		
		
	} else {
		if ($di eq $dirs[0]){
			print OUT "No Media Files Found<br>\n";
		} else {
			print COUT "No Media Files Found<br>\n";
		}	
		
	}
	}

	print OUT $HTML_FOOTER;
	if (defined $commondir and $dir ne $commondir){
		print COUT $HTML_FOOTER;
		close COUT;
	}	
	close OUT;

	
	
	
	
	
	
	
	


} # End sub report





#
# SUB get_document
# Accepts a URL, returns the source of the document at the URL
#
sub get_document { my ($self,$url,$dir,$commondir) = (shift,shift,shift,shift);		# Recieve as argument the URL to access
	my $ua = LWP::UserAgent->new;						# Create a new UserAgent
	$ua->agent('Mozilla/25.0');	# Give it a type name
	warn "Attempting to access <$url>...\n"  if $CHAT;
	if ($url =~ m|(\.$EXTENSIONS_BAD)$|i) {
		open LINK,">>$dir/LOGS/MEDIA.txt" or warn "Couldn't open to save <MEDIA>!" and return "Failed to save.";
			print LINK "$url\n";
		close LINK;
		if (defined $commondir and $dir ne $commondir){
			open LINK,">>$commondir/MEDIA.txt" or warn "Couldn't open to save <MEDIA>!" and return "Failed to save.";
			print LINK "$url\n";
			close LINK;
		}

		$url = '';

	} 
	my $req = new HTTP::Request('GET', $url) 			# Format URL request
		or warn "...could not GET.\n" and return undef;

	my $res = $ua->request($req);						# $res is the object UA returned
	if (not $res->is_success()) {							# If successful
		warn"...failed.\n"  if $CHAT;
		return undef
	}
	warn "...ok.\n" if $CHAT;
	# Test size
	if ($url =~ m|(\.$EXTENSIONS_RE)$|i) {
		$_ = length ($res->content);
		warn "Image file size $_ bytes\n" if $CHAT;
		if (exists $self->{MINIMGSIZE} and  $_ < $self->{MINIMGSIZE}){
			warn "Image size too small, ignoring.\n" if $CHAT;
			$self->{IGNORE}->{$url} = "Size $_ bytes is too small.";
			return undef;
		}
	}

	return $res->content;							# $res->content  is the HTML the UA returned from the URL
}



=head1 PRIVATE METHOD _save_img

Accepts the dir in which to store the image,
the image's URL (won't store same image twice)
and the actual image source. If file size is different from the last
time it was saved, saves both new and old
images to DIFFERENCES/, and runs $CMP
against them if set.

Returns the path the image was saved at.

=cut

sub _save_img { my ($self,$dir,$url,$img,$commondir) = (shift,shift,shift,shift,shift);
	local *OUT;
	my $filename;
	my @oldfile;
	my @lines;
	my $ofname;
	my $ext;
	# Remvoe any file path from the $url
	if (exists $self->{DONE}->{$url} or exists $self->{FAILED}->{$url}){
		warn "Already got this one ($url), not saving.\n" if $CHAT;
		return undef;
	}
	$url =~ m|/([^./]+)(\.$EXTENSIONS_RE)$|i;
	$ext = $2;
	if ($NEWNAMES == 1){
		$filename = $dir.'/'.(join'',localtime).$2;
	} elsif ($NEWNAMES == 2){
		$filename = $url;
		$filename =~ s/\/|\:|\~|\?/_/g;
		$filename = $dir.'/'.$filename;
	} else {
		$filename = "$dir/$1$2";
	}
	warn "Saving image as <$filename>...\n"  if $CHAT;
	@oldfile = stat($filename);
	if (exists $oldfile[7]){	#old file is there
		if (length($img) != $oldfile[7]){
			warn "Diff sizes: ".length($img)." vs $oldfile[7]" if $CHAT;
			$ofname = $filename;

			$filename = $url;
			$filename =~ s/\/|\:|\~/_/g;
			$filename = $dir.'/DIFFERENCES/'.$filename.'.'.(join'',localtime);
			$oldfile[5] = "$filename.old.$ext";		# temp fix
			$oldfile[5] =~ s/\//\\/g;
			$oldfile[4] = $ofname;		# temp fix
			$oldfile[4] =~ s/\//\\/g;
			
			$oldfile[6] = `move $oldfile[4] $oldfile[5]`;

			open OUT,">$ofname" or warn "Couldn't open to save <$ofname>!" and return "Failed to save.";
				binmode OUT;
				print OUT $img;
			close OUT;			

			open OUT,">$filename.new.$ext" or die "Couldn't open to save <$filename>!" and return "Failed to save.";
				binmode OUT;
				print OUT $img;
			close OUT;

			open OUT,">>$dir/LOGS/CHANGED.txt" or warn "Couldn't open to save <NEW>!" and return "Failed to save.";
				print OUT "$filename.old.$ext	$oldfile[7]\n";
				print OUT "$filename.new.$ext	".length($img)."\n";
			close OUT;
			if (defined $commondir and $dir ne $commondir){
				open OUT,">>$commondir/CHANGED.txt" or warn "Couldn't open to save <NEW>!" and return "Failed to save.";
				print OUT "$filename.old.$ext	$oldfile[7]\n";
				print OUT "$filename.new.$ext	".length($img)."\n";
				close OUT;
			}
			if ($CMP ne "NONE"){
				@lines = `$CMP -l $filename.old.$ext $filename.new.$ext`;
				open OUT,">>$dir/LOGS/CHANGED.txt" or warn "Couldn't open to save <NEW>!" and return "Failed to save.";
					print OUT "@lines\n\n";
				close OUT;				
				if (defined $commondir and $dir ne $commondir){
					open OUT,">>$commondir/CHANGED.txt" or warn "Couldn't open to save <NEW>!" and return "Failed to save.";
					print OUT "@lines\n\n";
					close OUT;
				}
			}
			
		}
	} else {			# this is a new file
		open OUT,">>$dir/LOGS/NEW.txt" or warn "Couldn't open to save <NEW>!" and return "Failed to save.";
			print OUT "$url\n";
		close OUT;
		if (defined $commondir and $dir ne $commondir){
				open OUT,">>$commondir/NEW.txt" or warn "Couldn't open to save <NEW>!" and return "Failed to save.";
				print OUT "$url\n";
				close OUT;
		}
		open OUT,">$filename" or warn "Couldn't open to save <$filename>!" and return "Failed to save.";
			binmode OUT;
			print OUT $img;
		close OUT;
	}
	warn "...image saved ok.\n" if $CHAT;
	return $filename;
}


#
# SUB abs_url returns an absolute URL for a $child_url linked from $parent_url
#
# DOC http://www.netverifier.com/pin/nicolette/jezfuzchr001.html
# SRC /pin/nicolette/jezfuzchr001.jpg
#
sub abs_url { my ($parent_url,$child_url) = (shift,shift);
	my $hack;
	if ($child_url =~ m|^/|) {
		$parent_url =~ s|^(http://[\w.]+)?/.*$|$1|i;
		return $parent_url.$child_url;
	}
	if ($child_url =~ m|^\.\.\/|i){
		$parent_url =~ s/\/[^\/]+$//;	# Strip filename
		warn $parent_url if $CHAT;
		while ($child_url=~s/^\.\.\///gs ){
			$parent_url =~s/[^\/|^~]+\/?$//;
#			if ($parent_url =~ /\/$/){$parent_url =~ s/\/$//;}
			warn $parent_url if $CHAT;
		}
		if ($child_url =~ /^\.\//){$child_url =~ s/^\.\///;}
		$child_url = $parent_url.$child_url;
	} elsif ($child_url !~ m/^http:\/\//i){
		# Assume relative path needs dir
		$parent_url =~ s/\/[^\/|^~]+$//;	# Strip filename
		if ($parent_url =~ /\/$/){$parent_url =~ s/\/$//;}
		if ($child_url =~ /^\.\//){$child_url =~ s/^\.\///;}
		$child_url = $parent_url .'/'.$child_url;
	}
	
	return $child_url;
}









1; # Return a true value for 'use'

=head1 SEE ALSO

Every thing and every one listed above under DEPENDENCIES.

=head1 SEE ACKNOWLEDGMENTS

The basis for the image-grabbing code is from HTTP::GetImages0.23, most of the bug fixes in StegTest should be in the next version of GetImages

=head1 AUTHOR

Duncan Lamb (duncan_lamb@hotmail.com) 12/23/2001 16:08

=head1 COPYRIGHT

Copyright 2001 Duncan Lamb.

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

